import enum
import os
import unittest
from collections.abc import Mapping, Sequence
from typing import Any

import pytest

import confuse

from . import _root


class ValidConfigTest(unittest.TestCase):
    def test_validate_simple_dict(self):
        config = _root({"foo": 5})
        valid = config.get({"foo": confuse.Integer()})
        assert valid["foo"] == 5

    def test_default_value(self):
        config = _root({})
        valid = config.get({"foo": confuse.Integer(8)})
        assert valid["foo"] == 8

    def test_undeclared_key_raises_keyerror(self):
        config = _root({"foo": 5})
        valid = config.get({"foo": confuse.Integer()})
        with pytest.raises(KeyError):
            valid["bar"]

    def test_undeclared_key_ignored_from_input(self):
        config = _root({"foo": 5, "bar": 6})
        valid = config.get({"foo": confuse.Integer()})
        with pytest.raises(KeyError):
            valid["bar"]

    def test_int_template_shortcut(self):
        config = _root({"foo": 5})
        valid = config.get({"foo": int})
        assert valid["foo"] == 5

    def test_int_default_shortcut(self):
        config = _root({})
        valid = config.get({"foo": 9})
        assert valid["foo"] == 9

    def test_attribute_access(self):
        config = _root({"foo": 5})
        valid = config.get({"foo": confuse.Integer()})
        assert valid.foo == 5

    def test_missing_required_value_raises_error_on_validate(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config.get({"foo": confuse.Integer()})

    def test_none_as_default(self):
        config = _root({})
        valid = config.get({"foo": confuse.Integer(None)})
        assert valid["foo"] is None

    def test_wrong_type_raises_error_on_validate(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigTypeError):
            config.get({"foo": confuse.Integer()})

    def test_validate_individual_value(self):
        config = _root({"foo": 5})
        valid = config["foo"].get(confuse.Integer())
        assert valid == 5

    def test_nested_dict_template(self):
        config = _root({"foo": {"bar": 9}})
        valid = config.get({"foo": {"bar": confuse.Integer()}})
        assert valid["foo"]["bar"] == 9

    def test_nested_attribute_access(self):
        config = _root({"foo": {"bar": 8}})
        valid = config.get({"foo": {"bar": confuse.Integer()}})
        assert valid.foo.bar == 8


class AsTemplateTest(unittest.TestCase):
    def test_plain_int_as_template(self):
        typ = confuse.as_template(int)
        assert isinstance(typ, confuse.Integer)
        assert typ.default == confuse.REQUIRED

    def test_concrete_int_as_template(self):
        typ = confuse.as_template(2)
        assert isinstance(typ, confuse.Integer)
        assert typ.default == 2

    def test_plain_string_as_template(self):
        typ = confuse.as_template(str)
        assert isinstance(typ, confuse.String)
        assert typ.default == confuse.REQUIRED

    def test_concrete_string_as_template(self):
        typ = confuse.as_template("foo")
        assert isinstance(typ, confuse.String)
        assert typ.default == "foo"

    def test_dict_as_template(self):
        typ = confuse.as_template({"key": 9})
        assert isinstance(typ, confuse.MappingTemplate)
        assert isinstance(typ.subtemplates["key"], confuse.Integer)
        assert typ.subtemplates["key"].default == 9

    def test_nested_dict_as_template(self):
        typ = confuse.as_template({"outer": {"inner": 2}})
        assert isinstance(typ, confuse.MappingTemplate)
        assert isinstance(typ.subtemplates["outer"], confuse.MappingTemplate)
        assert isinstance(
            typ.subtemplates["outer"].subtemplates["inner"], confuse.Integer
        )
        assert typ.subtemplates["outer"].subtemplates["inner"].default == 2

    def test_list_as_template(self):
        typ: confuse.OneOf[Any] = confuse.as_template(list())
        assert isinstance(typ, confuse.OneOf)
        assert typ.default == confuse.REQUIRED

    def test_set_as_template(self):
        typ: confuse.Choice[Any] = confuse.as_template(set())
        assert isinstance(typ, confuse.Choice)

    def test_enum_type_as_template(self):
        typ = confuse.as_template(enum.Enum)
        assert isinstance(typ, confuse.Choice)

    def test_float_type_as_tempalte(self):
        typ = confuse.as_template(float)
        assert isinstance(typ, confuse.Number)
        assert typ.default == confuse.REQUIRED

    def test_concrete_float_as_template(self):
        typ = confuse.as_template(2.0)
        assert isinstance(typ, confuse.Number)
        assert typ.default == 2.0

    def test_none_as_template(self):
        typ = confuse.as_template(None)
        assert type(typ) is confuse.Template
        assert typ.default is None

    def test_required_as_template(self):
        typ = confuse.as_template(confuse.REQUIRED)
        assert type(typ) is confuse.Template
        assert typ.default == confuse.REQUIRED

    def test_dict_type_as_template(self):
        typ = confuse.as_template(dict)
        assert isinstance(typ, confuse.TypeTemplate)
        assert typ.typ == Mapping
        assert typ.default == confuse.REQUIRED

    def test_list_type_as_template(self):
        typ = confuse.as_template(list)
        assert isinstance(typ, confuse.TypeTemplate)
        assert typ.typ == Sequence
        assert typ.default == confuse.REQUIRED

    def test_set_type_as_template(self):
        typ = confuse.as_template(set)
        assert isinstance(typ, confuse.TypeTemplate)
        assert typ.typ is set
        assert typ.default == confuse.REQUIRED

    def test_other_type_as_template(self):
        class MyClass:
            pass

        typ = confuse.as_template(MyClass)
        assert isinstance(typ, confuse.TypeTemplate)
        assert typ.typ == MyClass
        assert typ.default == confuse.REQUIRED


class StringTemplateTest(unittest.TestCase):
    def test_validate_string(self):
        config = _root({"foo": "bar"})
        valid = config.get({"foo": confuse.String()})
        assert valid["foo"] == "bar"

    def test_string_default_value(self):
        config = _root({})
        valid = config.get({"foo": confuse.String("baz")})
        assert valid["foo"] == "baz"

    def test_pattern_matching(self):
        config = _root({"foo": "bar", "baz": "zab"})
        valid = config.get({"foo": confuse.String(pattern="^ba.$")})
        assert valid["foo"] == "bar"
        with pytest.raises(confuse.ConfigValueError):
            config.get({"baz": confuse.String(pattern="!")})

    def test_string_template_shortcut(self):
        config = _root({"foo": "bar"})
        valid = config.get({"foo": str})
        assert valid["foo"] == "bar"

    def test_string_default_shortcut(self):
        config = _root({})
        valid = config.get({"foo": "bar"})
        assert valid["foo"] == "bar"

    def test_check_string_type(self):
        config = _root({"foo": 5})
        with pytest.raises(confuse.ConfigTypeError):
            config.get({"foo": confuse.String()})


class NumberTest(unittest.TestCase):
    def test_validate_int_as_number(self):
        config = _root({"foo": 2})
        valid = config["foo"].get(confuse.Number())
        assert isinstance(valid, int)
        assert valid == 2

    def test_validate_float_as_number(self):
        config = _root({"foo": 3.0})
        valid = config["foo"].get(confuse.Number())
        assert isinstance(valid, float)
        assert valid == 3.0

    def test_validate_string_as_number(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.Number())


class ChoiceTest(unittest.TestCase):
    def test_validate_good_choice_in_list(self):
        config = _root({"foo": 2})
        valid = config["foo"].get(confuse.Choice([1, 2, 4, 8, 16]))
        assert valid == 2

    def test_validate_bad_choice_in_list(self):
        config = _root({"foo": 3})
        with pytest.raises(confuse.ConfigValueError):
            config["foo"].get(confuse.Choice([1, 2, 4, 8, 16]))

    def test_validate_good_choice_in_dict(self):
        config = _root({"foo": 2})
        valid = config["foo"].get(confuse.Choice({2: "two", 4: "four"}))
        assert valid == "two"

    def test_validate_bad_choice_in_dict(self):
        config = _root({"foo": 3})
        with pytest.raises(confuse.ConfigValueError):
            config["foo"].get(confuse.Choice({2: "two", 4: "four"}))


class OneOfTest(unittest.TestCase):
    def test_default_value(self):
        config = _root({})
        valid = config["foo"].get(confuse.OneOf([], default="bar"))
        assert valid == "bar"

    def test_validate_good_choice_in_list(self):
        config = _root({"foo": 2})
        valid: str | int = config["foo"].get(
            confuse.OneOf(
                [
                    confuse.String(),
                    confuse.Integer(),
                ]
            )
        )
        assert valid == 2

    def test_validate_first_good_choice_in_list(self):
        config = _root({"foo": 3.14})
        valid: str | int = config["foo"].get(
            confuse.OneOf(
                [
                    confuse.Integer(),
                    confuse.Number(),
                ]
            )
        )
        assert valid == 3

    def test_validate_no_choice_in_list(self):
        config = _root({"foo": None})
        with pytest.raises(confuse.ConfigValueError):
            config["foo"].get(
                confuse.OneOf(
                    [
                        confuse.String(),
                        confuse.Integer(),
                    ]
                )
            )

    def test_validate_bad_template(self):
        class BadTemplate:
            pass

        config = _root({})
        with pytest.raises(ValueError, match="cannot convert to template"):
            config.get(confuse.OneOf([BadTemplate()]))
        del BadTemplate


class StrSeqTest(unittest.TestCase):
    def test_string_list(self):
        config = _root({"foo": ["bar", "baz"]})
        valid = config["foo"].get(confuse.StrSeq())
        assert valid == ["bar", "baz"]

    def test_string_tuple(self):
        config = _root({"foo": ("bar", "baz")})
        valid = config["foo"].get(confuse.StrSeq())
        assert valid == ["bar", "baz"]

    def test_whitespace_separated_string(self):
        config = _root({"foo": "bar   baz"})
        valid = config["foo"].get(confuse.StrSeq())
        assert valid == ["bar", "baz"]

    def test_invalid_type(self):
        config = _root({"foo": 9})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.StrSeq())

    def test_invalid_sequence_type(self):
        config = _root({"foo": ["bar", 2126]})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.StrSeq())


class FilenameTest(unittest.TestCase):
    def test_default_value(self):
        config = _root({})
        valid = config["foo"].get(confuse.Filename("foo/bar"))
        assert valid == "foo/bar"

    def test_default_none(self):
        config = _root({})
        valid = config["foo"].get(confuse.Filename(None))
        assert valid is None

    def test_missing_required_value(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.Filename())

    def test_filename_relative_to_working_dir(self):
        config = _root({"foo": "bar"})
        valid = config["foo"].get(confuse.Filename(cwd="/dev/null"))
        assert valid == os.path.realpath("/dev/null/bar")

    def test_filename_relative_to_sibling(self):
        config = _root({"foo": "/", "bar": "baz"})
        valid = config.get(
            {"foo": confuse.Filename(), "bar": confuse.Filename(relative_to="foo")}
        )
        assert valid.foo == os.path.realpath("/")
        assert valid.bar == os.path.realpath("/baz")

    def test_filename_working_dir_overrides_sibling(self):
        config = _root({"foo": "bar"})
        valid = config.get(
            {"foo": confuse.Filename(cwd="/dev/null", relative_to="baz")}
        )
        assert valid.foo == os.path.realpath("/dev/null/bar")

    def test_filename_relative_to_sibling_with_recursion(self):
        config = _root({"foo": "/", "bar": "r", "baz": "z"})
        with pytest.raises(confuse.ConfigTemplateError):
            config.get(
                {
                    "foo": confuse.Filename(relative_to="bar"),
                    "bar": confuse.Filename(relative_to="baz"),
                    "baz": confuse.Filename(relative_to="foo"),
                }
            )

    def test_filename_relative_to_self(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigTemplateError):
            config.get({"foo": confuse.Filename(relative_to="foo")})

    def test_filename_relative_to_sibling_needs_siblings(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigTemplateError):
            config["foo"].get(confuse.Filename(relative_to="bar"))

    def test_filename_relative_to_sibling_needs_template(self):
        config = _root({"foo": "/", "bar": "baz"})
        with pytest.raises(confuse.ConfigTemplateError):
            config.get({"bar": confuse.Filename(relative_to="foo")})

    def test_filename_with_non_file_source(self):
        config = _root({"foo": "foo/bar"})
        valid = config["foo"].get(confuse.Filename())
        assert valid == os.path.join(os.getcwd(), "foo", "bar")

    def test_filename_with_file_source(self):
        source = confuse.ConfigSource({"foo": "foo/bar"}, filename="/baz/config.yaml")
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename())
        assert valid == os.path.realpath("/config/path/foo/bar")

    def test_filename_with_default_source(self):
        source = confuse.ConfigSource(
            {"foo": "foo/bar"}, filename="/baz/config.yaml", default=True
        )
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename())
        assert valid == os.path.realpath("/config/path/foo/bar")

    def test_filename_use_config_source_dir(self):
        source = confuse.ConfigSource(
            {"foo": "foo/bar"}, filename="/baz/config.yaml", base_for_paths=True
        )
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename())
        assert valid == os.path.realpath("/baz/foo/bar")

    def test_filename_in_source_dir(self):
        source = confuse.ConfigSource({"foo": "foo/bar"}, filename="/baz/config.yaml")
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename(in_source_dir=True))
        assert valid == os.path.realpath("/baz/foo/bar")

    def test_filename_in_source_dir_overrides_in_app_dir(self):
        source = confuse.ConfigSource({"foo": "foo/bar"}, filename="/baz/config.yaml")
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename(in_source_dir=True, in_app_dir=True))
        assert valid == os.path.realpath("/baz/foo/bar")

    def test_filename_in_app_dir_non_file_source(self):
        source = confuse.ConfigSource({"foo": "foo/bar"})
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename(in_app_dir=True))
        assert valid == os.path.realpath("/config/path/foo/bar")

    def test_filename_in_app_dir_overrides_config_source_dir(self):
        source = confuse.ConfigSource(
            {"foo": "foo/bar"}, filename="/baz/config.yaml", base_for_paths=True
        )
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        valid = config["foo"].get(confuse.Filename(in_app_dir=True))
        assert valid == os.path.realpath("/config/path/foo/bar")

    def test_filename_wrong_type(self):
        config = _root({"foo": 8})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.Filename())


class PathTest(unittest.TestCase):
    def test_path_value(self):
        import pathlib

        config = _root({"foo": "foo/bar"})
        valid = config["foo"].get(confuse.Path())
        assert valid == pathlib.Path(os.path.abspath("foo/bar"))

    def test_default_value(self):
        import pathlib

        config = _root({})
        valid = config["foo"].get(confuse.Path("foo/bar"))
        assert valid == pathlib.Path("foo/bar")

    def test_default_none(self):
        config = _root({})
        valid = config["foo"].get(confuse.Path(None))
        assert valid is None

    def test_missing_required_value(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.Path())


class BaseTemplateTest(unittest.TestCase):
    def test_base_template_accepts_any_value(self):
        config = _root({"foo": 4.2})
        valid: float = config["foo"].get(confuse.Template())
        assert valid == 4.2

    def test_base_template_required(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.Template())

    def test_base_template_with_default(self):
        config = _root({})
        valid = config["foo"].get(confuse.Template("bar"))
        assert valid == "bar"


class TypeTemplateTest(unittest.TestCase):
    def test_correct_type(self):
        config = _root({"foo": set()})
        valid = config["foo"].get(confuse.TypeTemplate(set))
        assert valid == set()

    def test_incorrect_type(self):
        config = _root({"foo": dict()})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.TypeTemplate(set))

    def test_missing_required_value(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.TypeTemplate(set))

    def test_default_value(self):
        config = _root({})
        valid = config["foo"].get(confuse.TypeTemplate(set, {1, 2}))
        assert valid == {1, 2}


class SequenceTest(unittest.TestCase):
    def test_int_list(self):
        config = _root({"foo": [1, 2, 3]})
        valid = config["foo"].get(confuse.Sequence(int))
        assert valid == [1, 2, 3]

    def test_dict_list(self):
        config = _root({"foo": [{"bar": 1, "baz": 2}, {"bar": 3, "baz": 4}]})
        valid = config["foo"].get(confuse.Sequence({"bar": int, "baz": int}))
        assert valid == [{"bar": 1, "baz": 2}, {"bar": 3, "baz": 4}]

    def test_invalid_item(self):
        config = _root({"foo": [{"bar": 1, "baz": 2}, {"bar": 3, "bak": 4}]})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.Sequence({"bar": int, "baz": int}))

    def test_wrong_type(self):
        config = _root({"foo": {"one": 1, "two": 2, "three": 3}})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.Sequence(int))

    def test_missing(self):
        config = _root({"foo": [1, 2, 3]})
        valid = config["bar"].get(confuse.Sequence(int))
        assert valid == []


class MappingValuesTest(unittest.TestCase):
    def test_int_dict(self):
        config = _root({"foo": {"one": 1, "two": 2, "three": 3}})
        valid = config["foo"].get(confuse.MappingValues(int))
        assert valid == {"one": 1, "two": 2, "three": 3}

    def test_dict_dict(self):
        config = _root(
            {"foo": {"first": {"bar": 1, "baz": 2}, "second": {"bar": 3, "baz": 4}}}
        )
        valid: dict[str, dict[str, int]] = config["foo"].get(
            confuse.MappingValues({"bar": int, "baz": int})
        )
        assert valid == {"first": {"bar": 1, "baz": 2}, "second": {"bar": 3, "baz": 4}}

    def test_invalid_item(self):
        config = _root(
            {"foo": {"first": {"bar": 1, "baz": 2}, "second": {"bar": 3, "bak": 4}}}
        )
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.MappingValues({"bar": int, "baz": int}))

    def test_wrong_type(self):
        config = _root({"foo": [1, 2, 3]})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.MappingValues(int))

    def test_missing(self):
        config = _root({"foo": {"one": 1, "two": 2, "three": 3}})
        valid = config["bar"].get(confuse.MappingValues(int))
        assert valid == {}


class OptionalTest(unittest.TestCase):
    def test_optional_string_valid_type(self):
        config = _root({"foo": "bar"})
        valid = config["foo"].get(confuse.Optional(confuse.String()))
        assert valid == "bar"

    def test_optional_string_invalid_type(self):
        config = _root({"foo": 5})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(confuse.Optional(confuse.String()))

    def test_optional_string_null(self):
        config = _root({"foo": None})
        valid = config["foo"].get(confuse.Optional(confuse.String()))
        assert valid is None

    def test_optional_string_null_default_value(self):
        config = _root({"foo": None})
        valid = config["foo"].get(confuse.Optional(confuse.String(), "baz"))
        assert valid == "baz"

    def test_optional_string_null_string_provides_default(self):
        config = _root({"foo": None})
        valid = config["foo"].get(confuse.Optional(confuse.String("baz")))
        assert valid == "baz"

    def test_optional_string_null_string_default_override(self):
        config = _root({"foo": None})
        valid = config["foo"].get(
            confuse.Optional(confuse.String("baz"), default="bar")
        )
        assert valid == "bar"

    def test_optional_string_allow_missing_no_explicit_default(self):
        config = _root({})
        valid = config["foo"].get(confuse.Optional(confuse.String()))
        assert valid is None

    def test_optional_string_allow_missing_default_value(self):
        config = _root({})
        valid = config["foo"].get(confuse.Optional(confuse.String(), "baz"))
        assert valid == "baz"

    def test_optional_string_missing_not_allowed(self):
        config = _root({})
        with pytest.raises(confuse.NotFoundError):
            config["foo"].get(confuse.Optional(confuse.String(), allow_missing=False))

    def test_optional_string_null_missing_not_allowed(self):
        config = _root({"foo": None})
        valid = config["foo"].get(
            confuse.Optional(confuse.String(), allow_missing=False)
        )
        assert valid is None

    def test_optional_mapping_template_valid(self):
        config = _root({"foo": {"bar": 5, "baz": "bak"}})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template)})
        assert valid["foo"]
        assert valid["foo"]["bar"] == 5
        assert valid["foo"]["baz"] == "bak"

    def test_optional_mapping_template_invalid(self):
        config = _root({"foo": {"bar": 5, "baz": 10}})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        with pytest.raises(confuse.ConfigTypeError):
            config.get({"foo": confuse.Optional(template)})

    def test_optional_mapping_template_null(self):
        config = _root({"foo": None})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template)})
        assert valid["foo"] is None

    def test_optional_mapping_template_null_default_value(self):
        config = _root({"foo": None})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template, {})})
        assert isinstance(valid["foo"], dict)

    def test_optional_mapping_template_allow_missing_no_explicit_default(self):
        config = _root({})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template)})
        assert valid["foo"] is None

    def test_optional_mapping_template_allow_missing_default_value(self):
        config = _root({})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template, {})})
        assert isinstance(valid["foo"], dict)

    def test_optional_mapping_template_missing_not_allowed(self):
        config = _root({})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        with pytest.raises(confuse.NotFoundError):
            config.get({"foo": confuse.Optional(template, allow_missing=False)})

    def test_optional_mapping_template_null_missing_not_allowed(self):
        config = _root({"foo": None})
        template = {"bar": confuse.Integer(), "baz": confuse.String()}
        valid = config.get({"foo": confuse.Optional(template, allow_missing=False)})
        assert valid["foo"] is None



import os
import unittest
from unittest.mock import patch

import pytest

import confuse

from . import _root


class EnvSourceTest(unittest.TestCase):
    def setUp(self):
        self.env_patcher = patch.dict("os.environ", {})
        self.env_patcher.start()

    def tearDown(self):
        self.env_patcher.stop()

    def test_prefix(self):
        os.environ["TEST_FOO"] = "a"
        os.environ["BAR"] = "b"
        config = _root(confuse.EnvSource("TEST_"))
        assert config.get() == {"foo": "a"}

    def test_number_type_conversion(self):
        os.environ["TEST_FOO"] = "1"
        os.environ["TEST_BAR"] = "2.0"
        config = _root(confuse.EnvSource("TEST_"))
        foo = config["foo"].get()
        bar = config["bar"].get()
        assert isinstance(foo, int)
        assert foo == 1
        assert isinstance(bar, float)
        assert bar == 2.0

    def test_bool_type_conversion(self):
        os.environ["TEST_FOO"] = "true"
        os.environ["TEST_BAR"] = "FALSE"
        config = _root(confuse.EnvSource("TEST_"))
        assert config["foo"].get() is True
        assert config["bar"].get() is False

    def test_null_type_conversion(self):
        os.environ["TEST_FOO"] = "null"
        os.environ["TEST_BAR"] = ""
        config = _root(confuse.EnvSource("TEST_"))
        assert config["foo"].get() is None
        assert config["bar"].get() is None

    def test_unset_lower_config(self):
        os.environ["TEST_FOO"] = "null"
        config = _root({"foo": "bar"})
        assert config["foo"].get() == "bar"
        config.set(confuse.EnvSource("TEST_"))
        assert config["foo"].get() is None

    def test_sep_default(self):
        os.environ["TEST_FOO__BAR"] = "a"
        os.environ["TEST_FOO_BAZ"] = "b"
        config = _root(confuse.EnvSource("TEST_"))
        assert config["foo"]["bar"].get() == "a"
        assert config["foo_baz"].get() == "b"

    def test_sep_single_underscore_adjacent_seperators(self):
        os.environ["TEST_FOO__BAR"] = "a"
        os.environ["TEST_FOO_BAZ"] = "b"
        config = _root(confuse.EnvSource("TEST_", sep="_"))
        assert config["foo"][""]["bar"].get() == "a"
        assert config["foo"]["baz"].get() == "b"

    def test_nested(self):
        os.environ["TEST_FOO__BAR"] = "a"
        os.environ["TEST_FOO__BAZ__QUX"] = "b"
        config = _root(confuse.EnvSource("TEST_"))
        assert config["foo"]["bar"].get() == "a"
        assert config["foo"]["baz"]["qux"].get() == "b"

    def test_nested_rev(self):
        # Reverse to ensure order doesn't matter
        os.environ["TEST_FOO__BAZ__QUX"] = "b"
        os.environ["TEST_FOO__BAR"] = "a"
        config = _root(confuse.EnvSource("TEST_"))
        assert config["foo"]["bar"].get() == "a"
        assert config["foo"]["baz"]["qux"].get() == "b"

    def test_nested_clobber(self):
        os.environ["TEST_FOO__BAR"] = "a"
        os.environ["TEST_FOO__BAR__BAZ"] = "b"
        config = _root(confuse.EnvSource("TEST_"))
        # Clobbered
        assert config["foo"]["bar"].get() == {"baz": "b"}
        assert config["foo"]["bar"]["baz"].get() == "b"

    def test_nested_clobber_rev(self):
        # Reverse to ensure order doesn't matter
        os.environ["TEST_FOO__BAR__BAZ"] = "b"
        os.environ["TEST_FOO__BAR"] = "a"
        config = _root(confuse.EnvSource("TEST_"))
        # Clobbered
        assert config["foo"]["bar"].get() == {"baz": "b"}
        assert config["foo"]["bar"]["baz"].get() == "b"

    def test_lower_applied_after_prefix_match(self):
        os.environ["TEST_FOO"] = "a"
        config = _root(confuse.EnvSource("test_", lower=True))
        assert config.get() == {}

    def test_lower_already_lowercase(self):
        os.environ["TEST_foo"] = "a"
        config = _root(confuse.EnvSource("TEST_", lower=True))
        assert config.get() == {"foo": "a"}

    def test_lower_does_not_alter_value(self):
        os.environ["TEST_FOO"] = "UPPER"
        config = _root(confuse.EnvSource("TEST_", lower=True))
        assert config.get() == {"foo": "UPPER"}

    def test_lower_false(self):
        os.environ["TEST_FOO"] = "a"
        config = _root(confuse.EnvSource("TEST_", lower=False))
        assert config.get() == {"FOO": "a"}

    def test_handle_lists_good_list(self):
        os.environ["TEST_FOO__0"] = "a"
        os.environ["TEST_FOO__1"] = "b"
        os.environ["TEST_FOO__2"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == ["a", "b", "c"]

    def test_handle_lists_good_list_rev(self):
        # Reverse to ensure order doesn't matter
        os.environ["TEST_FOO__2"] = "c"
        os.environ["TEST_FOO__1"] = "b"
        os.environ["TEST_FOO__0"] = "a"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == ["a", "b", "c"]

    def test_handle_lists_nested_lists(self):
        os.environ["TEST_FOO__0__0"] = "a"
        os.environ["TEST_FOO__0__1"] = "b"
        os.environ["TEST_FOO__1__0"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == [["a", "b"], ["c"]]

    def test_handle_lists_bad_list_missing_index(self):
        os.environ["TEST_FOO__0"] = "a"
        os.environ["TEST_FOO__2"] = "b"
        os.environ["TEST_FOO__3"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == {"0": "a", "2": "b", "3": "c"}

    def test_handle_lists_bad_list_non_zero_start(self):
        os.environ["TEST_FOO__1"] = "a"
        os.environ["TEST_FOO__2"] = "b"
        os.environ["TEST_FOO__3"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == {"1": "a", "2": "b", "3": "c"}

    def test_handle_lists_bad_list_non_numeric(self):
        os.environ["TEST_FOO__0"] = "a"
        os.environ["TEST_FOO__ONE"] = "b"
        os.environ["TEST_FOO__2"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == {"0": "a", "one": "b", "2": "c"}

    def test_handle_lists_top_level_always_dict(self):
        os.environ["TEST_0"] = "a"
        os.environ["TEST_1"] = "b"
        os.environ["TEST_2"] = "c"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config.get() == {"0": "a", "1": "b", "2": "c"}

    def test_handle_lists_not_a_list(self):
        os.environ["TEST_FOO__BAR"] = "a"
        os.environ["TEST_FOO__BAZ"] = "b"
        config = _root(confuse.EnvSource("TEST_", handle_lists=True))
        assert config["foo"].get() == {"bar": "a", "baz": "b"}

    def test_parse_yaml_docs_scalar(self):
        os.environ["TEST_FOO"] = "a"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"].get() == "a"

    def test_parse_yaml_docs_list(self):
        os.environ["TEST_FOO"] = "[a, b]"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"].get() == ["a", "b"]

    def test_parse_yaml_docs_dict(self):
        os.environ["TEST_FOO"] = "{bar: a, baz: b}"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"].get() == {"bar": "a", "baz": "b"}

    def test_parse_yaml_docs_nested(self):
        os.environ["TEST_FOO"] = "{bar: [a, b], baz: {qux: c}}"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"]["bar"].get() == ["a", "b"]
        assert config["foo"]["baz"].get() == {"qux": "c"}

    def test_parse_yaml_docs_number_conversion(self):
        os.environ["TEST_FOO"] = "{bar: 1, baz: 2.0}"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        bar = config["foo"]["bar"].get()
        baz = config["foo"]["baz"].get()
        assert isinstance(bar, int)
        assert bar == 1
        assert isinstance(baz, float)
        assert baz == 2.0

    def test_parse_yaml_docs_bool_conversion(self):
        os.environ["TEST_FOO"] = "{bar: true, baz: FALSE}"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"]["bar"].get() is True
        assert config["foo"]["baz"].get() is False

    def test_parse_yaml_docs_null_conversion(self):
        os.environ["TEST_FOO"] = "{bar: null, baz: }"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))
        assert config["foo"]["bar"].get() is None
        assert config["foo"]["baz"].get() is None

    def test_parse_yaml_docs_syntax_error(self):
        os.environ["TEST_FOO"] = "{:}"
        with pytest.raises(confuse.ConfigError, match="TEST_FOO"):
            _root(confuse.EnvSource("TEST_", parse_yaml_docs=True))

    def test_parse_yaml_docs_false(self):
        os.environ["TEST_FOO"] = "{bar: a, baz: b}"
        config = _root(confuse.EnvSource("TEST_", parse_yaml_docs=False))
        assert config["foo"].get() == "{bar: a, baz: b}"
        with pytest.raises(confuse.ConfigError):
            config["foo"]["bar"].get()


class ConfigEnvTest(unittest.TestCase):
    def setUp(self):
        self.env_patcher = patch.dict(
            "os.environ",
            {
                "TESTAPP_FOO": "a",
                "TESTAPP_BAR__NESTED": "b",
                "TESTAPP_BAZ_SEP_NESTED": "c",
                "MYAPP_QUX_SEP_NESTED": "d",
            },
        )
        self.env_patcher.start()
        self.config = confuse.Configuration("TestApp", read=False)

    def test_defaults(self):
        self.config.set_env()
        assert self.config.get() == {
            "foo": "a",
            "bar": {"nested": "b"},
            "baz_sep_nested": "c",
        }

    def test_with_prefix(self):
        self.config.set_env(prefix="MYAPP_")
        assert self.config.get() == {"qux_sep_nested": "d"}

    def test_with_sep(self):
        self.config.set_env(sep="_sep_")
        assert self.config.get() == {
            "foo": "a",
            "bar__nested": "b",
            "baz": {"nested": "c"},
        }

    def test_with_prefix_and_sep(self):
        self.config.set_env(prefix="MYAPP_", sep="_sep_")
        assert self.config.get() == {"qux": {"nested": "d"}}



import datetime as dt
import os
import re
import sys
from pathlib import Path

MATCH_VERSION_LINE = re.compile(r"version = \W((\d+\.\d+)\.\d.*?)\W").fullmatch

pyproject = Path(__file__).parent.parent / "pyproject.toml"
version_line_match = next(
    filter(None, map(MATCH_VERSION_LINE, pyproject.read_text().splitlines()))
)
release, version = version_line_match.groups()

sys.path.insert(0, os.path.abspath(".."))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinx_rtd_theme",
]
source_suffix = ".rst"
master_doc = "index"

project = "Confuse"
copyright = f"2012-{dt.date.today().year}, Adrian Sampson & contributors"

exclude_patterns = ["_build"]

pygments_style = "sphinx"

# -- Options for HTML output --------------------------------------------------

html_theme = "sphinx_rtd_theme"
htmlhelp_basename = "Confusedoc"



import textwrap
import unittest
from collections import OrderedDict

import confuse

from . import _root


class PrettyDumpTest(unittest.TestCase):
    def test_dump_null(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": None})
        yaml = config.dump().strip()
        assert yaml == "foo:"

    def test_dump_true(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": True})
        yaml = config.dump().strip()
        assert yaml == "foo: yes"

    def test_dump_false(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": False})
        yaml = config.dump().strip()
        assert yaml == "foo: no"

    def test_dump_short_list(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": ["bar", "baz"]})
        yaml = config.dump().strip()
        assert yaml == "foo: [bar, baz]"

    def test_dump_ordered_dict(self):
        odict = OrderedDict()
        odict["foo"] = "bar"
        odict["bar"] = "baz"
        odict["baz"] = "qux"

        config = confuse.Configuration("myapp", read=False)
        config.add({"key": odict})
        yaml = config.dump().strip()
        assert (
            yaml
            == textwrap.dedent("""
            key:
                foo: bar
                bar: baz
                baz: qux
        """).strip()
        )

    def test_dump_sans_defaults(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": "bar"})
        config.sources[0].default = True
        config.add({"baz": "qux"})

        yaml = config.dump().strip()
        assert yaml == "foo: bar\nbaz: qux"

        yaml = config.dump(full=False).strip()
        assert yaml == "baz: qux"


class RedactTest(unittest.TestCase):
    def test_no_redaction(self):
        config = _root({"foo": "bar"})
        data = config.flatten(redact=True)
        assert data == {"foo": "bar"}

    def test_redact_key(self):
        config = _root({"foo": "bar"})
        config["foo"].redact = True
        data = config.flatten(redact=True)
        assert data == {"foo": "REDACTED"}

    def test_unredact(self):
        config = _root({"foo": "bar"})
        config["foo"].redact = True
        config["foo"].redact = False
        data = config.flatten(redact=True)
        assert data == {"foo": "bar"}

    def test_dump_redacted(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": "bar"})
        config["foo"].redact = True
        yaml = config.dump(redact=True).strip()
        assert yaml == "foo: REDACTED"

    def test_dump_unredacted(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": "bar"})
        config["foo"].redact = True
        yaml = config.dump(redact=False).strip()
        assert yaml == "foo: bar"

    def test_dump_redacted_sans_defaults(self):
        config = confuse.Configuration("myapp", read=False)
        config.add({"foo": "bar"})
        config.sources[0].default = True
        config.add({"baz": "qux"})
        config["baz"].redact = True

        yaml = config.dump(redact=True, full=False).strip()
        assert yaml == "baz: REDACTED"



import unittest
from argparse import Namespace
from collections import OrderedDict
from typing import Any

import pytest

import confuse


class BuildDictTests(unittest.TestCase):
    def test_pure_dicts(self):
        config = {"foo": {"bar": 1}}
        result = confuse.util.build_dict(config)
        assert 1 == result["foo"]["bar"]

    def test_namespaces(self):
        config = Namespace(foo=Namespace(bar=2), another=1)
        result = confuse.util.build_dict(config)
        assert 2 == result["foo"]["bar"]
        assert 1 == result["another"]

    def test_dot_sep_keys(self):
        config: dict[str, Any] = {"foo.bar": 1}
        result = confuse.util.build_dict(config.copy())
        assert 1 == result["foo.bar"]

        result = confuse.util.build_dict(config.copy(), sep=".")
        assert 1 == result["foo"]["bar"]

    def test_dot_sep_keys_clobber(self):
        args: list[tuple[str, Any]] = [("foo.bar", 1), ("foo.bar.zar", 2)]
        config = OrderedDict(args)
        result = confuse.util.build_dict(config.copy(), sep=".")
        assert {"zar": 2} == result["foo"]["bar"]
        assert 2 == result["foo"]["bar"]["zar"]

        # Reverse and do it again! (should be stable)
        args.reverse()
        config = OrderedDict(args)
        result = confuse.util.build_dict(config.copy(), sep=".")
        assert {"zar": 2} == result["foo"]["bar"]
        assert 2 == result["foo"]["bar"]["zar"]

    def test_dot_sep_keys_no_clobber(self):
        args: list[tuple[str, Any]] = [
            ("foo.bar", 1),
            ("foo.far", 2),
            ("foo.zar.dar", 4),
        ]
        config = OrderedDict(args)
        result = confuse.util.build_dict(config.copy(), sep=".")
        assert 1 == result["foo"]["bar"]
        assert 2 == result["foo"]["far"]
        assert 4 == result["foo"]["zar"]["dar"]

    def test_adjacent_underscores_sep_keys(self):
        config: dict[str, Any] = {"foo__bar_baz": 1}
        result = confuse.util.build_dict(config.copy())
        assert 1 == result["foo__bar_baz"]

        result = confuse.util.build_dict(config.copy(), sep="_")
        assert 1 == result["foo"][""]["bar"]["baz"]

        result = confuse.util.build_dict(config.copy(), sep="__")
        assert 1 == result["foo"]["bar_baz"]

    def test_keep_none(self):
        config = {"foo": None}
        result = confuse.util.build_dict(config.copy())
        with pytest.raises(KeyError):
            result["foo"]

        result = confuse.util.build_dict(config.copy(), keep_none=True)
        assert None is result["foo"]

    def test_keep_none_with_nested(self):
        config = {"foo": {"bar": None}}
        result = confuse.util.build_dict(config.copy())
        assert {} == result["foo"]

        result = confuse.util.build_dict(config.copy(), keep_none=True)
        assert None is result["foo"]["bar"]



from __future__ import annotations

import unittest
from typing import TYPE_CHECKING

import pytest

import confuse

from . import _root

if TYPE_CHECKING:
    from collections.abc import Iterator

    from confuse import Subview


class SingleSourceTest(unittest.TestCase):
    def test_dict_access(self):
        config = _root({"foo": "bar"})
        value = config["foo"].get()
        assert value == "bar"

    def test_list_access(self):
        config = _root({"foo": ["bar", "baz"]})
        value = config["foo"][1].get()
        assert value == "baz"

    def test_missing_key(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.NotFoundError):
            config["baz"].get()

    def test_missing_index(self):
        config = _root({"l": ["foo", "bar"]})
        with pytest.raises(confuse.NotFoundError):
            config["l"][5].get()

    def test_dict_iter(self):
        config = _root({"foo": "bar", "baz": "qux"})
        keys = [key for key in config]
        assert set(keys) == {"foo", "baz"}

    def test_list_iter(self):
        config = _root({"l": ["foo", "bar"]})
        # TODO(@snejus): we need to split Subview to SequenceView and MappingView in
        # order to have automatic resolution here
        _items: Iterator[Subview] = iter(config["l"])
        items = [subview.get() for subview in _items]
        assert items == ["foo", "bar"]

    def test_int_iter(self):
        config = _root({"n": 2})
        with pytest.raises(confuse.ConfigTypeError):
            [item for item in config["n"]]

    def test_dict_keys(self):
        config = _root({"foo": "bar", "baz": "qux"})
        keys = config.keys()
        assert set(keys) == {"foo", "baz"}

    def test_dict_values(self):
        config = _root({"foo": "bar", "baz": "qux"})
        values = [value.get() for value in config.values()]
        assert set(values) == {"bar", "qux"}

    def test_dict_items(self):
        config = _root({"foo": "bar", "baz": "qux"})
        items = [(key, value.get()) for (key, value) in config.items()]
        assert set(items) == {("foo", "bar"), ("baz", "qux")}

    def test_list_keys_error(self):
        config = _root({"l": ["foo", "bar"]})
        with pytest.raises(confuse.ConfigTypeError):
            config["l"].keys()

    def test_list_sequence(self):
        config = _root({"l": ["foo", "bar"]})
        items = [item.get() for item in config["l"].sequence()]
        assert items == ["foo", "bar"]

    def test_dict_sequence_error(self):
        config = _root({"foo": "bar", "baz": "qux"})
        with pytest.raises(confuse.ConfigTypeError):
            list(config.sequence())

    def test_dict_contents(self):
        config = _root({"foo": "bar", "baz": "qux"})
        contents = config.all_contents()
        assert set(contents) == {"foo", "baz"}

    def test_list_contents(self):
        config = _root({"l": ["foo", "bar"]})
        contents = config["l"].all_contents()
        assert list(contents) == ["foo", "bar"]

    def test_int_contents(self):
        config = _root({"n": 2})
        with pytest.raises(confuse.ConfigTypeError):
            list(config["n"].all_contents())


class ConverstionTest(unittest.TestCase):
    def test_str_conversion_from_str(self):
        config = _root({"foo": "bar"})
        value = str(config["foo"])
        assert value == "bar"

    def test_str_conversion_from_int(self):
        config = _root({"foo": 2})
        value = str(config["foo"])
        assert value == "2"

    def test_bool_conversion_from_bool(self):
        config = _root({"foo": True})
        value = bool(config["foo"])
        assert value

    def test_bool_conversion_from_int(self):
        config = _root({"foo": 0})
        value = bool(config["foo"])
        assert not value


class NameTest(unittest.TestCase):
    def test_root_name(self):
        config = _root()
        assert config.name == "root"

    def test_string_access_name(self):
        config = _root()
        name = config["foo"].name
        assert name == "foo"

    def test_int_access_name(self):
        config = _root()
        name = config[5].name
        assert name == "#5"

    def test_nested_access_name(self):
        config = _root()
        name = config[5]["foo"]["bar"][20].name
        assert name == "#5.foo.bar#20"


class MultipleSourceTest(unittest.TestCase):
    def test_dict_access_shadowed(self):
        config = _root({"foo": "bar"}, {"foo": "baz"})
        value = config["foo"].get()
        assert value == "bar"

    def test_dict_access_fall_through(self):
        config = _root({"qux": "bar"}, {"foo": "baz"})
        value = config["foo"].get()
        assert value == "baz"

    def test_dict_access_missing(self):
        config = _root({"qux": "bar"}, {"foo": "baz"})
        with pytest.raises(confuse.NotFoundError):
            config["fred"].get()

    def test_list_access_shadowed(self):
        config = _root({"l": ["a", "b"]}, {"l": ["c", "d", "e"]})
        value = config["l"][1].get()
        assert value == "b"

    def test_list_access_fall_through(self):
        config = _root({"l": ["a", "b"]}, {"l": ["c", "d", "e"]})
        value = config["l"][2].get()
        assert value == "e"

    def test_list_access_missing(self):
        config = _root({"l": ["a", "b"]}, {"l": ["c", "d", "e"]})
        with pytest.raises(confuse.NotFoundError):
            config["l"][3].get()

    def test_access_dict_replaced(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"qux": "fred"}})
        value = config["foo"].get()
        assert value == {"bar": "baz"}

    def test_dict_keys_merged(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"qux": "fred"}})
        keys = config["foo"].keys()
        assert set(keys) == {"bar", "qux"}

    def test_dict_keys_replaced(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"bar": "fred"}})
        keys = config["foo"].keys()
        assert list(keys) == ["bar"]

    def test_dict_values_merged(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"qux": "fred"}})
        values = [value.get() for value in config["foo"].values()]
        assert set(values) == {"baz", "fred"}

    def test_dict_values_replaced(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"bar": "fred"}})
        values = [value.get() for value in config["foo"].values()]
        assert list(values) == ["baz"]

    def test_dict_items_merged(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"qux": "fred"}})
        items = [(key, value.get()) for (key, value) in config["foo"].items()]
        assert set(items) == {("bar", "baz"), ("qux", "fred")}

    def test_dict_items_replaced(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"bar": "fred"}})
        items = [(key, value.get()) for (key, value) in config["foo"].items()]
        assert list(items) == [("bar", "baz")]

    def test_list_sequence_shadowed(self):
        config = _root({"l": ["a", "b"]}, {"l": ["c", "d", "e"]})
        items = [item.get() for item in config["l"].sequence()]
        assert items == ["a", "b"]

    def test_list_sequence_shadowed_by_dict(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": ["qux", "fred"]})
        with pytest.raises(confuse.ConfigTypeError):
            list(config["foo"].sequence())

    def test_dict_contents_concatenated(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"qux": "fred"}})
        contents = config["foo"].all_contents()
        assert set(contents) == {"bar", "qux"}

    def test_dict_contents_concatenated_not_replaced(self):
        config = _root({"foo": {"bar": "baz"}}, {"foo": {"bar": "fred"}})
        contents = config["foo"].all_contents()
        assert list(contents) == ["bar", "bar"]

    def test_list_contents_concatenated(self):
        config = _root({"foo": ["bar", "baz"]}, {"foo": ["qux", "fred"]})
        contents = config["foo"].all_contents()
        assert list(contents) == ["bar", "baz", "qux", "fred"]

    def test_int_contents_error(self):
        config = _root({"foo": ["bar", "baz"]}, {"foo": 5})
        with pytest.raises(confuse.ConfigTypeError):
            list(config["foo"].all_contents())

    def test_list_and_dict_contents_concatenated(self):
        config = _root({"foo": ["bar", "baz"]}, {"foo": {"qux": "fred"}})
        contents = config["foo"].all_contents()
        assert list(contents) == ["bar", "baz", "qux"]

    def test_add_source(self):
        config = _root({"foo": "bar"})
        config.add({"baz": "qux"})
        assert config["foo"].get() == "bar"
        assert config["baz"].get() == "qux"


class SetTest(unittest.TestCase):
    def test_set_missing_top_level_key(self):
        config = _root({})
        config["foo"] = "bar"
        assert config["foo"].get() == "bar"

    def test_override_top_level_key(self):
        config = _root({"foo": "bar"})
        config["foo"] = "baz"
        assert config["foo"].get() == "baz"

    def test_set_second_level_key(self):
        config = _root({})
        config["foo"]["bar"] = "baz"
        assert config["foo"]["bar"].get() == "baz"

    def test_override_second_level_key(self):
        config = _root({"foo": {"bar": "qux"}})
        config["foo"]["bar"] = "baz"
        assert config["foo"]["bar"].get() == "baz"

    def test_override_list_index(self):
        config = _root({"foo": ["a", "b", "c"]})
        config["foo"][1] = "bar"
        assert config["foo"][1].get() == "bar"



import os
import shutil
import tempfile
from typing import Any

import confuse


def _root(*sources: dict[str, Any]) -> confuse.RootView:
    return confuse.RootView([confuse.ConfigSource.of(s) for s in sources])


class TempDir:
    """Context manager that creates and destroys a temporary directory."""

    def __init__(self):
        self.path = tempfile.mkdtemp()

    def __enter__(self):
        return self

    def __exit__(self, *errstuff):
        shutil.rmtree(self.path)

    def sub(self, name, contents=None):
        """Get a path to a file named `name` inside this temporary
        directory. If `contents` is provided, then the bytestring is
        written to the file.
        """
        path = os.path.join(self.path, name)
        if contents:
            with open(path, "wb") as f:
                f.write(contents)
        return path



import argparse
import optparse
import unittest
from argparse import Namespace

import pytest

import confuse


class ArgparseTest(unittest.TestCase):
    def setUp(self):
        self.config = confuse.Configuration("test", read=False)
        self.parser = argparse.ArgumentParser()

    def _parse(self, args, **kwargs):
        args = self.parser.parse_args(args.split())
        self.config.set_args(args, **kwargs)

    def test_text_argument_parsed(self):
        self.parser.add_argument("--foo", metavar="BAR")
        self._parse("--foo bar")
        assert self.config["foo"].get() == "bar"

    def test_boolean_argument_parsed(self):
        self.parser.add_argument("--foo", action="store_true")
        self._parse("--foo")
        assert self.config["foo"].get()

    def test_missing_optional_argument_not_included(self):
        self.parser.add_argument("--foo", metavar="BAR")
        self._parse("")
        with pytest.raises(confuse.NotFoundError):
            self.config["foo"].get()

    def test_argument_overrides_default(self):
        self.config.add({"foo": "baz"})

        self.parser.add_argument("--foo", metavar="BAR")
        self._parse("--foo bar")
        assert self.config["foo"].get() == "bar"

    def test_nested_destination_single(self):
        self.parser.add_argument("--one", dest="one.foo")
        self.parser.add_argument("--two", dest="one.two.foo")
        self._parse("--two TWO", dots=True)
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_nested(self):
        self.parser.add_argument("--one", dest="one.foo")
        self.parser.add_argument("--two", dest="one.two.foo")
        self._parse("--two TWO --one ONE", dots=True)
        assert self.config["one"]["foo"].get() == "ONE"
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_nested_rev(self):
        self.parser.add_argument("--one", dest="one.foo")
        self.parser.add_argument("--two", dest="one.two.foo")
        # Reverse to ensure order doesn't matter
        self._parse("--one ONE --two TWO", dots=True)
        assert self.config["one"]["foo"].get() == "ONE"
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_clobber(self):
        self.parser.add_argument("--one", dest="one.two")
        self.parser.add_argument("--two", dest="one.two.foo")
        self._parse("--two TWO --one ONE", dots=True)
        # Clobbered
        assert self.config["one"]["two"].get() == {"foo": "TWO"}
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_clobber_rev(self):
        # Reversed order
        self.parser.add_argument("--two", dest="one.two.foo")
        self.parser.add_argument("--one", dest="one.two")
        self._parse("--one ONE --two TWO", dots=True)
        # Clobbered just the same
        assert self.config["one"]["two"].get() == {"foo": "TWO"}
        assert self.config["one"]["two"]["foo"].get() == "TWO"


class OptparseTest(unittest.TestCase):
    def setUp(self):
        self.config = confuse.Configuration("test", read=False)
        self.parser = optparse.OptionParser()

    def _parse(self, args, **kwargs):
        options, _ = self.parser.parse_args(args.split())
        self.config.set_args(options, **kwargs)

    def test_text_argument_parsed(self):
        self.parser.add_option("--foo", metavar="BAR")
        self._parse("--foo bar")
        assert self.config["foo"].get() == "bar"

    def test_boolean_argument_parsed(self):
        self.parser.add_option("--foo", action="store_true")
        self._parse("--foo")
        assert self.config["foo"].get()

    def test_missing_optional_argument_not_included(self):
        self.parser.add_option("--foo", metavar="BAR")
        self._parse("")
        with pytest.raises(confuse.NotFoundError):
            self.config["foo"].get()

    def test_argument_overrides_default(self):
        self.config.add({"foo": "baz"})

        self.parser.add_option("--foo", metavar="BAR")
        self._parse("--foo bar")
        assert self.config["foo"].get() == "bar"

    def test_nested_destination_single(self):
        self.parser.add_option("--one", dest="one.foo")
        self.parser.add_option("--two", dest="one.two.foo")
        self._parse("--two TWO", dots=True)
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_nested(self):
        self.parser.add_option("--one", dest="one.foo")
        self.parser.add_option("--two", dest="one.two.foo")
        self._parse("--two TWO --one ONE", dots=True)
        assert self.config["one"]["foo"].get() == "ONE"
        assert self.config["one"]["two"]["foo"].get() == "TWO"

    def test_nested_destination_nested_rev(self):
        self.parser.add_option("--one", dest="one.foo")
        self.parser.add_option("--two", dest="one.two.foo")
        # Reverse to ensure order doesn't matter
        self._parse("--one ONE --two TWO", dots=True)
        assert self.config["one"]["foo"].get() == "ONE"
        assert self.config["one"]["two"]["foo"].get() == "TWO"


class GenericNamespaceTest(unittest.TestCase):
    def setUp(self):
        self.config = confuse.Configuration("test", read=False)

    def test_value_added_to_root(self):
        self.config.set_args(Namespace(foo="bar"))
        assert self.config["foo"].get() == "bar"

    def test_value_added_to_subview(self):
        self.config["baz"].set_args(Namespace(foo="bar"))
        assert self.config["baz"]["foo"].get() == "bar"

    def test_nested_namespace(self):
        args = Namespace(first="Hello", nested=Namespace(second="World"))
        self.config.set_args(args, dots=True)
        assert self.config["first"].get() == "Hello"
        assert self.config["nested"]["second"].get() == "World"



import ntpath
import os
import platform
import posixpath
import shutil
import tempfile
import unittest
from typing import ClassVar

import confuse
import confuse.yaml_util

DEFAULT = (platform.system, os.environ, os.path)
SYSTEMS = {
    "Linux": ({"HOME": "/home/test", "XDG_CONFIG_HOME": "~/xdgconfig"}, posixpath),
    "Darwin": ({"HOME": "/Users/test"}, posixpath),
    "Windows": (
        {
            "APPDATA": "~\\winconfig",
            "HOME": "C:\\Users\\test",
            "USERPROFILE": "C:\\Users\\test",
        },
        ntpath,
    ),
}


def _touch(path):
    open(path, "a").close()


class FakeHome(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.home = tempfile.mkdtemp()
        os.environ["HOME"] = self.home
        os.environ["USERPROFILE"] = self.home

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.home)


class FakeSystem(unittest.TestCase):
    SYS_NAME: ClassVar[str]

    def setUp(self):
        super().setUp()
        self.os_path = os.path
        os.environ = {}  # type: ignore[assignment]

        environ, os.path = SYSTEMS[self.SYS_NAME]
        os.environ.update(environ)
        platform.system = lambda: self.SYS_NAME

    def tearDown(self):
        super().tearDown()
        platform.system, os.environ, os.path = DEFAULT


class LinuxTestCases(FakeSystem):
    SYS_NAME = "Linux"

    def test_both_xdg_and_fallback_dirs(self):
        assert confuse.config_dirs() == [
            "/home/test/.config",
            "/home/test/xdgconfig",
            "/etc/xdg",
            "/etc",
        ]

    def test_fallback_only(self):
        del os.environ["XDG_CONFIG_HOME"]
        assert confuse.config_dirs() == ["/home/test/.config", "/etc/xdg", "/etc"]

    def test_xdg_matching_fallback_not_duplicated(self):
        os.environ["XDG_CONFIG_HOME"] = "~/.config"
        assert confuse.config_dirs() == ["/home/test/.config", "/etc/xdg", "/etc"]

    def test_xdg_config_dirs(self):
        os.environ["XDG_CONFIG_DIRS"] = "/usr/local/etc/xdg:/etc/xdg"
        assert confuse.config_dirs() == [
            "/home/test/.config",
            "/home/test/xdgconfig",
            "/usr/local/etc/xdg",
            "/etc/xdg",
            "/etc",
        ]


class OSXTestCases(FakeSystem):
    SYS_NAME = "Darwin"

    def test_mac_dirs(self):
        assert confuse.config_dirs() == [
            "/Users/test/.config",
            "/Users/test/Library/Application Support",
            "/etc/xdg",
            "/etc",
        ]

    def test_xdg_config_dirs(self):
        os.environ["XDG_CONFIG_DIRS"] = "/usr/local/etc/xdg:/etc/xdg"
        assert confuse.config_dirs() == [
            "/Users/test/.config",
            "/Users/test/Library/Application Support",
            "/usr/local/etc/xdg",
            "/etc/xdg",
            "/etc",
        ]


class WindowsTestCases(FakeSystem):
    SYS_NAME = "Windows"

    def test_dir_from_environ(self):
        assert confuse.config_dirs() == [
            "C:\\Users\\test\\AppData\\Roaming",
            "C:\\Users\\test\\winconfig",
        ]

    def test_fallback_dir(self):
        del os.environ["APPDATA"]
        assert confuse.config_dirs() == ["C:\\Users\\test\\AppData\\Roaming"]


class ConfigFilenamesTest(unittest.TestCase):
    def setUp(self):
        self._old = os.path.isfile, confuse.yaml_util.load_yaml
        os.path.isfile = lambda x: True  # type: ignore[assignment]
        confuse.yaml_util.load_yaml = lambda *args, **kwargs: {}

    def tearDown(self):
        os.path.isfile, confuse.yaml_util.load_yaml = self._old

    def test_no_sources_when_files_missing(self):
        config = confuse.Configuration("myapp", read=False)
        filenames = [s.filename for s in config.sources]
        assert filenames == []

    def test_search_package(self):
        config = confuse.Configuration("myapp", __name__, read=False)
        config._add_default_source()

        for source in config.sources:
            if source.default:
                default_source = source
                break
        else:
            self.fail("no default source")

        assert default_source.filename == os.path.join(
            os.path.dirname(__file__), "config_default.yaml"
        )
        assert source.default


class EnvVarTest(FakeHome):
    def setUp(self):
        super().setUp()
        self.config = confuse.Configuration("myapp", read=False)
        os.environ["MYAPPDIR"] = self.home  # use the tmp home as a config dir

    def test_env_var_name(self):
        assert self.config._env_var == "MYAPPDIR"

    def test_env_var_dir_has_first_priority(self):
        assert self.config.config_dir() == self.home

    def test_env_var_missing(self):
        del os.environ["MYAPPDIR"]
        assert self.config.config_dir() != self.home


@unittest.skipUnless(platform.system() == "Linux", "Linux-specific tests")
class PrimaryConfigDirTest(FakeHome, FakeSystem):
    SYS_NAME = "Linux"  # conversion from posix to nt is easy

    def setUp(self):
        super().setUp()

        self.config = confuse.Configuration("test", read=False)

    def test_create_dir_if_none_exists(self):
        path = os.path.join(self.home, ".config", "test")
        assert not os.path.exists(path)

        assert self.config.config_dir() == path
        assert os.path.isdir(path)

    def test_return_existing_dir(self):
        path = os.path.join(self.home, "xdgconfig", "test")
        os.makedirs(path)
        _touch(os.path.join(path, confuse.CONFIG_FILENAME))
        assert self.config.config_dir() == path

    def test_do_not_create_dir_if_lower_priority_exists(self):
        path1 = os.path.join(self.home, "xdgconfig", "test")
        path2 = os.path.join(self.home, ".config", "test")
        os.makedirs(path2)
        _touch(os.path.join(path2, confuse.CONFIG_FILENAME))
        assert not os.path.exists(path1)
        assert os.path.exists(path2)

        assert self.config.config_dir() == path2
        assert not os.path.isdir(path1)
        assert os.path.isdir(path2)



"""An example application using Confuse for configuration."""

import argparse

import confuse

template = {
    "library": confuse.Filename(),
    "import_write": confuse.OneOf([bool, "ask", "skip"]),
    "ignore": confuse.StrSeq(),
    "plugins": list,
    "paths": {
        "directory": confuse.Filename(),
        "default": confuse.Filename(relative_to="directory"),
    },
    "servers": confuse.Sequence(
        {
            "hostname": str,
            "options": confuse.StrSeq(),
        }
    ),
}

config = confuse.LazyConfig("ConfuseExample", __name__)


def main() -> None:
    parser = argparse.ArgumentParser(description="example Confuse program")
    parser.add_argument(
        "--library",
        "-l",
        dest="library",
        metavar="LIBPATH",
        help="library database file",
    )
    parser.add_argument(
        "--directory",
        "-d",
        dest="paths.directory",
        metavar="DIRECTORY",
        help="destination music directory",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        dest="verbose",
        action="store_true",
        help="print debugging messages",
    )

    args = parser.parse_args()
    config.set_args(args, dots=True)

    print("configuration directory is", config.config_dir())

    # Use a boolean flag and the transient overlay.
    if config["verbose"]:
        print("verbose mode")
        config["log"]["level"] = 2
    else:
        config["log"]["level"] = 0
    print("logging level is", config["log"]["level"].get(int))

    valid = config.get(template)

    # Some validated/converted values.
    print("library is", valid.library)
    print("directory is", valid.paths.directory)
    print("paths.default is", valid.paths.default)
    print("servers are", [s.hostname for s in valid.servers])



import enum
import os
import unittest

import pytest

import confuse

from . import _root


class TypeCheckTest(unittest.TestCase):
    def test_str_type_correct(self):
        config = _root({"foo": "bar"})
        value = config["foo"].get(str)
        assert value == "bar"

    def test_str_type_incorrect(self):
        config = _root({"foo": 2})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(str)

    def test_int_type_correct(self):
        config = _root({"foo": 2})
        value = config["foo"].get(int)
        assert value == 2

    def test_int_type_incorrect(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].get(int)


class BuiltInValidatorTest(unittest.TestCase):
    def test_as_filename_with_non_file_source(self):
        config = _root({"foo": "foo/bar"})
        value = config["foo"].as_filename()
        assert value == os.path.join(os.getcwd(), "foo", "bar")

    def test_as_filename_with_file_source(self):
        source = confuse.ConfigSource({"foo": "foo/bar"}, filename="/baz/config.yaml")
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        value = config["foo"].as_filename()
        assert value == os.path.realpath("/config/path/foo/bar")

    def test_as_filename_with_default_source(self):
        source = confuse.ConfigSource(
            {"foo": "foo/bar"}, filename="/baz/config.yaml", default=True
        )
        config = _root(source)
        config.config_dir = lambda: "/config/path"  # type: ignore[attr-defined]
        value = config["foo"].as_filename()
        assert value == os.path.realpath("/config/path/foo/bar")

    def test_as_filename_wrong_type(self):
        config = _root({"foo": None})
        with pytest.raises(confuse.ConfigTypeError):
            config["foo"].as_filename()

    def test_as_path(self):
        config = _root({"foo": "foo/bar"})
        path_str = os.path.join(os.getcwd(), "foo", "bar")
        try:
            import pathlib
        except ImportError:
            with pytest.raises(ImportError):
                value = config["foo"].as_path()
        else:
            value = config["foo"].as_path()
            path = pathlib.Path(path_str)
            assert value == path

    def test_as_choice_correct(self):
        config = _root({"foo": "bar"})
        value = config["foo"].as_choice(["foo", "bar", "baz"])
        assert value == "bar"

    def test_as_choice_error(self):
        config = _root({"foo": "bar"})
        with pytest.raises(confuse.ConfigValueError):
            config["foo"].as_choice(["foo", "baz"])

    def test_as_choice_with_dict(self):
        config = _root({"foo": "bar"})
        res = config["foo"].as_choice(
            {
                "bar": "baz",
                "x": "y",
            }
        )
        assert res == "baz"

    def test_as_choice_with_enum(self):
        class Foobar(enum.Enum):
            Foo = "bar"

        config = _root({"foo": Foobar.Foo.value})
        res = config["foo"].as_choice(Foobar)
        assert res == Foobar.Foo

    def test_as_choice_with_enum_error(self):
        class Foobar(enum.Enum):
            Foo = "bar"

        config = _root({"foo": "foo"})
        with pytest.raises(confuse.ConfigValueError):
            config["foo"].as_choice(Foobar)

    def test_as_number_float(self):
        config = _root({"f": 1.0})
        config["f"].as_number()

    def test_as_number_int(self):
        config = _root({"i": 2})
        config["i"].as_number()

    def test_as_number_string(self):
        config = _root({"s": "a"})
        with pytest.raises(confuse.ConfigTypeError):
            config["s"].as_number()

    def test_as_str_seq_str(self):
        config = _root({"k": "a b c"})
        assert config["k"].as_str_seq() == ["a", "b", "c"]

    def test_as_str_seq_list(self):
        config = _root({"k": ["a b", "c"]})
        assert config["k"].as_str_seq() == ["a b", "c"]

    def test_as_str(self):
        config = _root({"s": "foo"})
        config["s"].as_str()

    def test_as_str_non_string(self):
        config = _root({"f": 1.0})
        with pytest.raises(confuse.ConfigTypeError):
            config["f"].as_str()

    def test_as_str_expanded(self):
        config = _root({"s": "${CONFUSE_TEST_VAR}/bar"})
        os.environ["CONFUSE_TEST_VAR"] = "foo"
        assert config["s"].as_str_expanded() == "foo/bar"

    def test_as_pairs(self):
        config = _root({"k": [{"a": "A"}, "b", ["c", "C"]]})
        assert [("a", "A"), ("b", None), ("c", "C")] == config["k"].as_pairs()



import unittest
from collections import OrderedDict

import pytest
import yaml

import confuse

from . import TempDir


def load(s):
    return yaml.load(s, Loader=confuse.Loader)


class ParseTest(unittest.TestCase):
    def test_dict_parsed_as_ordereddict(self):
        v = load("a: b\nc: d")
        assert isinstance(v, OrderedDict)
        assert list(v) == ["a", "c"]

    def test_string_beginning_with_percent(self):
        v = load("foo: %bar")
        assert v["foo"] == "%bar"


class FileParseTest(unittest.TestCase):
    def _parse_contents(self, contents):
        with TempDir() as temp:
            path = temp.sub("test_config.yaml", contents)
            return confuse.load_yaml(path)

    def test_load_file(self):
        v = self._parse_contents(b"foo: bar")
        assert v["foo"] == "bar"

    def test_syntax_error(self):
        with pytest.raises(confuse.ConfigError, match=r"test_config\.yaml"):
            self._parse_contents(b":")

    def test_reload_conf(self):
        with TempDir() as temp:
            path = temp.sub("test_config.yaml", b"foo: bar")
            config = confuse.Configuration("test", __name__)
            config.set_file(filename=path)
            assert config["foo"].get() == "bar"
            temp.sub("test_config.yaml", b"foo: bar2\ntest: hello world")
            config.reload()
            assert config["foo"].get() == "bar2"
            assert config["test"].get() == "hello world"

    def test_tab_indentation_error(self):
        with pytest.raises(confuse.ConfigError, match="found tab"):
            self._parse_contents(b"foo:\n\tbar: baz")


class StringParseTest(unittest.TestCase):
    def test_load_string(self):
        v = confuse.load_yaml_string("foo: bar", "test")
        assert v["foo"] == "bar"

    def test_string_syntax_error(self):
        with pytest.raises(confuse.ConfigError, match="test"):
            confuse.load_yaml_string(":", "test")

    def test_string_tab_indentation_error(self):
        with pytest.raises(confuse.ConfigError, match="found tab"):
            confuse.load_yaml_string("foo:\n\tbar: baz", "test")


class ParseAsScalarTest(unittest.TestCase):
    def test_text_string(self):
        v = confuse.yaml_util.parse_as_scalar("foo", confuse.Loader)
        assert v == "foo"

    def test_number_string_to_int(self):
        v = confuse.yaml_util.parse_as_scalar("1", confuse.Loader)
        assert isinstance(v, int)
        assert v == 1

    def test_number_string_to_float(self):
        v = confuse.yaml_util.parse_as_scalar("1.0", confuse.Loader)
        assert isinstance(v, float)
        assert v == 1.0

    def test_bool_string_to_bool(self):
        v = confuse.yaml_util.parse_as_scalar("true", confuse.Loader)
        assert v is True

    def test_empty_string_to_none(self):
        v = confuse.yaml_util.parse_as_scalar("", confuse.Loader)
        assert v is None

    def test_null_string_to_none(self):
        v = confuse.yaml_util.parse_as_scalar("null", confuse.Loader)
        assert v is None

    def test_dict_string_unchanged(self):
        v = confuse.yaml_util.parse_as_scalar('{"foo": "bar"}', confuse.Loader)
        assert v == '{"foo": "bar"}'

    def test_dict_unchanged(self):
        v = confuse.yaml_util.parse_as_scalar({"foo": "bar"}, confuse.Loader)
        assert v == {"foo": "bar"}

    def test_list_string_unchanged(self):
        v = confuse.yaml_util.parse_as_scalar('["foo", "bar"]', confuse.Loader)
        assert v == '["foo", "bar"]'

    def test_list_unchanged(self):
        v = confuse.yaml_util.parse_as_scalar(["foo", "bar"], confuse.Loader)
        assert v == ["foo", "bar"]

    def test_invalid_yaml_string_unchanged(self):
        v = confuse.yaml_util.parse_as_scalar("!", confuse.Loader)
        assert v == "!"
